/*
 * Copyright (c) 1991-1993 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the Computer Systems
 *	Engineering Group at Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char rcsid[] =
    "@(#) $Header: sun_audio.cc,v 1.20 95/10/16 22:47:49 van Exp $ (LBL)";
#endif

#undef resource_h

#include <stropts.h>
#include "audio-ibm.h"

#include <sys/errno.h>
#include <fcntl.h>
#include <strings.h>

static class IBMAudioMatcher : public Matcher {
    public:
	IBMAudioMatcher() : Matcher("audio") {}
	TclObject* match(const char* id) {
		if (strcasecmp(id, "ibm") == 0)
			return create();
		return (0);
	}
	Audio* create();
} IBMaudio_matcher;

Audio* IBMAudioMatcher::create()
{
	Tcl& tcl = Tcl::instance();
	const char* device = tcl.attr("audioFileName");
	int i;
	printf("try to create from %s\n", device);
	return (new IBMAudio(device));
}

IBMAudio::IBMAudio(const char* device)
{
    printf("create %s blksize=%d\n", device, blksize);

    iports = 2;
    oports = 2;

    sprintf(rec_device, "%s/1", device);
    sprintf(play_device, "%s/2", device);

    bufcur = buf = new u_char[blksize];
    bufend = buf + blksize;
}

int IBMAudio::OPort(int p)
{
    printf("oport %d\n", p);

    switch(p) {
    default:
    case output_speaker:
	return EXTERNAL_SPEAKER;
    case output_phones:
	return EXTERNAL_SPEAKER;
    case output_line:
	return OUTPUT_1;
    }
}

int IBMAudio::IPort(int p)
{
    printf("iport %d\n", p);

    switch(p)
    {
    default:
    case input_mike:
/*	return HIGH_GAIN_MIKE; */
	return LOW_GAIN_MIKE;
    case input_line:
	return LINE_1;
    }
}

void IBMAudio::Obtain()
{
    printf("obtain\n");
    if (!HaveAudio())
    {
	int flags;
	flags = O_RDONLY; // | O_NDELAY;
	printf("open '%s' flags=%x\n", rec_device, flags);
	rec_fd = ::open(rec_device, flags);
	if (rec_fd < 0)
	{
	    perror("rec_fd open");
	    exit(1);
	}
	printf("open %s\n", play_device);
	play_fd = open(play_device, O_WRONLY | O_NDELAY);
	if (play_fd < 0)
	{
	    perror("play_fd open");
	    exit(1);
	}
	fd = rec_fd;
	setup_device();
	Audio::Obtain();
	start();
    }
}

void IBMAudio::start()
{
    audio_control control;

    control.ioctl_request =	AUDIO_START;
    control.request_info =	NULL;
    control.position =	0;
    control.return_code =	0xDeadBeaf;

    if (ioctl(rec_fd, AUDIO_CONTROL, &control) < 0)
    {
	perror("ioctl AUDIO_CONTROL start");
	exit(1);
    }

    control.ioctl_request =	AUDIO_START;
    control.request_info =	NULL;
    control.position =	0;
    control.return_code =	0xDeadBeaf;

    if (ioctl(play_fd, AUDIO_CONTROL, &control) < 0)
    {
	perror("ioctl AUDIO_CONTROL start");
	exit(1);
    }
}

void IBMAudio::stop()
{
    audio_control control;

    control.ioctl_request =	AUDIO_STOP;
    control.request_info =	NULL;
    control.position =	0;
    control.return_code =	0xDeadBeaf;

    if (ioctl(rec_fd, AUDIO_CONTROL, &control) < 0)
    {
	perror("ioctl AUDIO_CONTROL stop");
	exit(1);
    }
}

void IBMAudio::Release()
{
    printf("release\n");

    if (HaveAudio())
    {
	stop();
	close(play_fd);
	play_fd = -1;
	Audio::Release();
    }
}

void IBMAudio::Write(u_char *cp)
{
//    printf("write %d\n", blksize);
    
    register int len = blksize;

    int cc = write(play_fd, (char *)cp, len);

    if ((len -= cc) != 0)
    {
	do {
	    if (cc < 0) {
//		if (errno != EPERM)
		    perror("audio write");
		return;
	    }
	    cp += cc;
	    cc = write(play_fd, (char *)cp, len);
	    if (cc < 0) {
		if (errno != EPERM)
		    perror("audio write");
		return;
	    }
	    len -= cc;
	} while (len > 0);
    }
}

int IBMAudio::FrameReady()
{

    register int len = bufend - bufcur;
//    printf("frameReady len=%d\n", len);
    while (len > 0)
    {
	int cc = read(rec_fd, (char *)bufcur, len);
//	printf("Read len=%d returns %d\n", len, cc);
	if (cc < 0)
	{
	    printf("Took error %d\n", errno);
	    switch (errno) {

	    case EWOULDBLOCK:
	    case EBUSY:
	    case 0:
	    case 192:
	    case 160:
	    case 224:
		/* should warn about audio_bsize here */
		goto out;

	    default:
		perror("audio read");
		Release();
		Obtain();
		return (0);
	    }
	}
	bufcur += cc;
	len -= cc;
    }
 out:
    return (bufcur >= bufend);
}

u_char* IBMAudio::Read()
{
//    printf("read\n");
    bufcur = buf;
    return (buf);
}

void IBMAudio::setup_device()
{
    audio_init init;

    bzero(&init, sizeof(init));
    init.srate = 8000;
    init.bits_per_sample = 8;
    init.bsize = blksize;
    init.mode = MU_LAW;
    init.channels = 1;
    init.position_resolution = AUDIO_IGNORE;
    init.flags = FIXED;
    init.operation = RECORD;
    init.reserved = NULL;

    if (ioctl(rec_fd, AUDIO_INIT, &init) < 0)
    {
	perror("rec_fd ioctl AUDIO_INIT");
	exit(1);
    }

    bzero(&init, sizeof(init));
    init.srate = 8000;
    init.bits_per_sample = 8;
    init.bsize = blksize;
    init.mode = MU_LAW;
    init.channels = 1;
    init.position_resolution = AUDIO_IGNORE;
    init.flags = FIXED;
    init.operation = PLAY;
    init.reserved = NULL;

    if (ioctl(play_fd, AUDIO_INIT, &init) < 0)
    {
	perror("rec_fd ioctl AUDIO_INIT");
	exit(1);
    }

    set_monitor(0);
    set_capture_buffer(4);
    set_output(OPort(oport));
    set_input(IPort(iport));

    bufcur = buf;
}

void IBMAudio::set_capture_buffer(int t)
{
    init_buf_req bufreq;

    bufreq.buffer_size = 8000* t;	/* has to be multiple of #bytes/s */
    bufreq.play_block_count = AUDIO_IGNORE;
    bufreq.play_lower_limit = AUDIO_IGNORE;
    bufreq.cap_upper_limit = AUDIO_IGNORE;
    bufreq.byte_count = AUDIO_IGNORE;
    bufreq.return_code = 0xDeadBeaf;

    if (ioctl (rec_fd, AUDIO_MODIFY_LIMITS, &bufreq) < 0)
    {
	perror("ioctl AUDIO_MODIFY_LIMITS");
    }
}

void IBMAudio::set_monitor(int flag)
{
    audio_change change;
    audio_control control;

    init_change(&change);
    init_control(&control);

    if (flag)
	change.monitor = MONITOR_UNCOMPRESSED;
    else
	change.monitor = MONITOR_OFF;

    control.request_info = &change;
    control.ioctl_request = AUDIO_CHANGE;

    if (ioctl(rec_fd, AUDIO_CONTROL, &control) < 0)
    {
	perror("ioctl rec_fd AUDIO_CONTROL monitor");
    }
}

void IBMAudio::init_change(audio_change *c)
{
    c->dev_info =	NULL;
    c->input =		AUDIO_IGNORE;
    c->output =		AUDIO_IGNORE;
    c->monitor =	AUDIO_IGNORE;
    c->volume =		AUDIO_IGNORE;
    c->volume_delay =	AUDIO_IGNORE;
    c->balance =	AUDIO_IGNORE;
    c->balance_delay =	AUDIO_IGNORE;
    c->treble =		AUDIO_IGNORE;		/* not supported */
    c->bass =		AUDIO_IGNORE;		/* not supported */
    c->pitch =		AUDIO_IGNORE;		/* not supported */
}

void IBMAudio::init_control(audio_control *c)
{
    c->ioctl_request =	AUDIO_IGNORE;
    c->request_info =	NULL;
    c->position =	0;
    c->return_code =	0xDeadBeaf;
}

int IBMAudio::GainClip(int level)
{
    printf("gainclip %d\n", level);
#if 0
        if (level < AUDIO_MIN_GAIN)
                return AUDIO_MIN_GAIN;
        else if (level > AUDIO_MAX_GAIN)
                return AUDIO_MAX_GAIN;
        else
#endif
                return level;

}


//
// Record gain
//
void IBMAudio::SetRGain(int level)
{
    audio_set_gain gain;
    
    printf("setrgain %d\n", level);

    level--;
    if (level < 0)
	level = 0;
    if (level > 255)
	level = 255;

    gain.right_gain = gain.left_gain = level * 100 /255;

    if (ioctl(rec_fd, AUDIO_SET_GAIN, &gain) < 0)
    {
	perror("ioctl AUDIO_SET_GAIN rec_fd");
    }
}

//
// playback gain
//
void IBMAudio::SetPGain(int level)
{
    audio_change change;
    audio_control control;
    
    printf("setpgain %d\n", level);

    level--;
    if (level < 0)
	level = 0;
    if (level > 255)
	level = 255;

    init_change(&change);
    init_control(&control);
	
    change.volume = 0x7fff0000 / 255 * level;
    change.volume_delay = 0;
    control.request_info = &change;
    control.ioctl_request = AUDIO_CHANGE;
    if (ioctl (play_fd, AUDIO_CONTROL, &control) == -1)
    {
	perror("ioctl play_fd AUDIO_CONTROL change_volume");
    }
}

void IBMAudio::OutputPort(int p)
{
    printf("outputport %d\n", p);
    oport = p;

    set_output(OPort(p));
}

void IBMAudio::InputPort(int p)
{
    printf("InputPort %d\n", p);
    iport = p;

    set_input(IPort(p));
}

void IBMAudio::set_input(int port)
{
    audio_change change;
    audio_control control;
    printf("set_input %d\n", port);

    init_change(&change);
    init_control(&control);

    change.input = port;
    control.request_info = &change;
    control.ioctl_request = AUDIO_CHANGE;

    if (ioctl (rec_fd, AUDIO_CONTROL, &control) < 0)
    {
	perror("ioctl rec_fd AUDIO_CONTROL set_input");
    }
}

void IBMAudio::set_output(int port)
{
    audio_change change;
    audio_control control;
    printf("set_output %d\n", port);

    init_change(&change);
    init_control(&control);

    change.output = port;
    control.request_info = &change;
    control.ioctl_request = AUDIO_CHANGE;

    if (ioctl (rec_fd, AUDIO_CONTROL, &control) < 0)
    {
	perror("ioctl rec_fd AUDIO_CONTROL set_output");
    }
}


extern const short mulawtolin[];
extern const unsigned char lintomulaw[];
extern const unsigned char lintomulawX[];

